from dataclasses import dataclass, field
from typing import ClassVar, NoReturn

from fastapi import HTTPException, status


@dataclass
class HTTPExcRaiser:
    message: str | None = field(default=None, kw_only=True)
    status_code: int | None = field(default=None, kw_only=True)

    default_exc_message: ClassVar[str] = "Permission denied"
    default_exc_status_code: ClassVar[int] = status.HTTP_403_FORBIDDEN

    def get_exc_message(self) -> str:
        return self.message or self.default_exc_message

    def get_exc_status_code(self) -> int:
        return self.status_code or self.default_exc_status_code

    def raise_http_exception(self, message: str | None) -> NoReturn:
        raise HTTPException(
            status_code=self.get_exc_status_code(),
            detail=message or self.get_exc_message(),
        )


__all__ = [
    "HTTPExcRaiser",
]



import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


def _identity_hash(self: Any) -> int:
    return id(self)


# This is a workaround to add a default __hash__ method to dataclasses that don't define one.
# We need this because dataclasses are not checking for __hash__ method in base classes.
#
# In general, we need to add __hash__ method cause it required by fastapi's dependency injection system
# for object to be hashable to be used as a dependency.
class IdentityHashMixin:
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        if kwargs.get("no_hash_override", False):
            return

        cls.__hash__ = _identity_hash  # type: ignore[assignment]


@dataclass
class SignatureOverride(ABC):
    def __post_init__(self) -> None:
        self.__signature__ = self.__get_signature__()

    @abstractmethod
    def __get_signature__(self) -> inspect.Signature:
        pass


__all__ = [
    "IdentityHashMixin",
    "SignatureOverride",
]



from collections.abc import Iterator
from dataclasses import dataclass
from typing import Annotated

import pytest
from fastapi import Depends, FastAPI, Header
from fastapi.exceptions import RequestValidationError
from fastapi.testclient import TestClient

from fastapi_has_permissions import Permission, lazy

app = FastAPI()


@dataclass
class AgeIsMoreThan(Permission):
    age: int

    async def check_permissions(self, age: Annotated[int, Header()]) -> bool:
        return age > self.age


@app.get(
    "/age-restricted-endpoint",
    dependencies=[
        Depends(lazy(AgeIsMoreThan(age=18), skip_on_exc=(RequestValidationError,))),
    ],
)
async def route() -> str:
    return "You have access to this endpoint!"


@pytest.fixture(scope="session")
def app_client() -> Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


@pytest.mark.parametrize(
    ("endpoint", "headers", "expected_status"),
    [
        pytest.param("/age-restricted-endpoint", {"age": "20"}, 200, id="age-over-18"),
        pytest.param("/age-restricted-endpoint", {"age": "17"}, 403, id="age-under-18"),
        pytest.param("/age-restricted-endpoint", {"age": "invalid"}, 200, id="invalid-age"),
        pytest.param("/age-restricted-endpoint", {}, 200, id="missing-age"),
    ],
)
def test_permissions(endpoint, headers, expected_status, app_client) -> None:
    response = app_client.get(endpoint, headers=headers)
    assert response.status_code == expected_status



from collections.abc import Iterator
from typing import Annotated

import pytest
from fastapi import Depends, FastAPI, Header, status
from fastapi.testclient import TestClient

from fastapi_has_permissions.common import HasRole, HasScope, IsAuthenticated


async def get_is_authenticated(authorization: Annotated[str | None, Header()] = None) -> bool:
    return authorization is not None


async def get_scopes(x_scopes: Annotated[str | None, Header()] = None) -> list[str]:
    if x_scopes is None:
        return []

    return x_scopes.split(",")


async def get_role(x_role: Annotated[str | None, Header()] = None) -> str:
    return x_role or ""


app = FastAPI()


@app.get(
    "/is-authenticated",
    dependencies=[
        Depends(
            IsAuthenticated(
                Depends(get_is_authenticated),
            ),
        ),
    ],
)
async def route() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/has-scope-read",
    dependencies=[
        Depends(
            HasScope(
                Depends(get_scopes),
                scopes=["read"],
            ),
        ),
    ],
)
async def route_has_scope_read() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/has-scope-read-write",
    dependencies=[
        Depends(
            HasScope(
                Depends(get_scopes),
                scopes=["read", "write"],
            ),
        ),
    ],
)
async def route_has_scope_read_write() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/has-role-admin",
    dependencies=[
        Depends(
            HasRole(
                Depends(get_role),
                roles=["admin"],
            ),
        ),
    ],
)
async def route_has_role_admin() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/has-role-admin-or-moderator",
    dependencies=[
        Depends(
            HasRole(
                Depends(get_role),
                roles=["admin", "moderator"],
            ),
        ),
    ],
)
async def route_has_role_admin_or_moderator() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/authenticated-and-admin",
    dependencies=[
        Depends(
            IsAuthenticated(Depends(get_is_authenticated)) & HasRole(Depends(get_role), roles=["admin"]),
        ),
    ],
)
async def route_authenticated_and_admin() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/admin-or-moderator-role",
    dependencies=[
        Depends(
            HasRole(Depends(get_role), roles=["admin"]) | HasRole(Depends(get_role), roles=["moderator"]),
        ),
    ],
)
async def route_admin_or_moderator_role() -> str:
    return "You have access to this endpoint!"


@app.get(
    "/not-admin",
    dependencies=[
        Depends(
            ~HasRole(Depends(get_role), roles=["admin"]),
        ),
    ],
)
async def route_not_admin() -> str:
    return "You have access to this endpoint!"


@pytest.fixture(scope="session")
def app_client() -> Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


@pytest.mark.parametrize(
    ("endpoint", "headers", "expected_status"),
    [
        pytest.param(
            "/is-authenticated",
            {"Authorization": "Bearer token"},
            status.HTTP_200_OK,
            id="is-authenticated-pass",
        ),
        pytest.param(
            "/is-authenticated",
            {},
            status.HTTP_403_FORBIDDEN,
            id="is-authenticated-fail-no-header",
        ),
        pytest.param(
            "/has-scope-read",
            {"x-scopes": "read"},
            status.HTTP_200_OK,
            id="has-scope-read-pass",
        ),
        pytest.param(
            "/has-scope-read",
            {"x-scopes": "write"},
            status.HTTP_403_FORBIDDEN,
            id="has-scope-read-fail-wrong-scope",
        ),
        pytest.param(
            "/has-scope-read",
            {},
            status.HTTP_403_FORBIDDEN,
            id="has-scope-read-fail-no-scopes",
        ),
        pytest.param(
            "/has-scope-read",
            {"x-scopes": "read,write"},
            status.HTTP_200_OK,
            id="has-scope-read-pass-multiple-scopes-including-read",
        ),
        pytest.param(
            "/has-scope-read-write",
            {"x-scopes": "read,write"},
            status.HTTP_200_OK,
            id="has-scope-read-write-pass-both",
        ),
        pytest.param(
            "/has-scope-read-write",
            {"x-scopes": "read"},
            status.HTTP_403_FORBIDDEN,
            id="has-scope-read-write-fail-missing-write",
        ),
        pytest.param(
            "/has-scope-read-write",
            {"x-scopes": "write"},
            status.HTTP_403_FORBIDDEN,
            id="has-scope-read-write-fail-missing-read",
        ),
        pytest.param(
            "/has-scope-read-write",
            {},
            status.HTTP_403_FORBIDDEN,
            id="has-scope-read-write-fail-no-scopes",
        ),
        pytest.param(
            "/has-scope-read-write",
            {"x-scopes": "read,write,delete"},
            status.HTTP_200_OK,
            id="has-scope-read-write-pass-superset-of-required",
        ),
        # HasRole - single role
        pytest.param(
            "/has-role-admin",
            {"x-role": "admin"},
            status.HTTP_200_OK,
            id="has-role-admin-pass",
        ),
        pytest.param(
            "/has-role-admin",
            {"x-role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="has-role-admin-fail-wrong-role",
        ),
        pytest.param(
            "/has-role-admin",
            {},
            status.HTTP_403_FORBIDDEN,
            id="has-role-admin-fail-no-role",
        ),
        pytest.param(
            "/has-role-admin-or-moderator",
            {"x-role": "admin"},
            status.HTTP_200_OK,
            id="has-role-admin-or-moderator-pass-admin",
        ),
        pytest.param(
            "/has-role-admin-or-moderator",
            {"x-role": "moderator"},
            status.HTTP_200_OK,
            id="has-role-admin-or-moderator-pass-moderator",
        ),
        pytest.param(
            "/has-role-admin-or-moderator",
            {"x-role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="has-role-admin-or-moderator-fail-wrong-role",
        ),
        pytest.param(
            "/has-role-admin-or-moderator",
            {},
            status.HTTP_403_FORBIDDEN,
            id="has-role-admin-or-moderator-fail-no-role",
        ),
        pytest.param(
            "/authenticated-and-admin",
            {"Authorization": "Bearer token", "x-role": "admin"},
            status.HTTP_200_OK,
            id="authenticated-and-admin-pass",
        ),
        pytest.param(
            "/authenticated-and-admin",
            {"Authorization": "Bearer token", "x-role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="authenticated-and-admin-fail-wrong-role",
        ),
        pytest.param(
            "/authenticated-and-admin",
            {"x-role": "admin"},
            status.HTTP_403_FORBIDDEN,
            id="authenticated-and-admin-fail-not-authenticated",
        ),
        pytest.param(
            "/authenticated-and-admin",
            {},
            status.HTTP_403_FORBIDDEN,
            id="authenticated-and-admin-fail-neither",
        ),
        pytest.param(
            "/admin-or-moderator-role",
            {"x-role": "admin"},
            status.HTTP_200_OK,
            id="admin-or-moderator-pass-admin",
        ),
        pytest.param(
            "/admin-or-moderator-role",
            {"x-role": "moderator"},
            status.HTTP_200_OK,
            id="admin-or-moderator-pass-moderator",
        ),
        pytest.param(
            "/admin-or-moderator-role",
            {"x-role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="admin-or-moderator-fail-wrong-role",
        ),
        pytest.param(
            "/admin-or-moderator-role",
            {},
            status.HTTP_403_FORBIDDEN,
            id="admin-or-moderator-fail-no-role",
        ),
        pytest.param(
            "/not-admin",
            {"x-role": "user"},
            status.HTTP_200_OK,
            id="not-admin-pass-non-admin-role",
        ),
        pytest.param(
            "/not-admin",
            {},
            status.HTTP_200_OK,
            id="not-admin-pass-no-role",
        ),
        pytest.param(
            "/not-admin",
            {"x-role": "admin"},
            status.HTTP_403_FORBIDDEN,
            id="not-admin-fail-is-admin",
        ),
    ],
)
def test_permissions(endpoint, headers, expected_status, app_client) -> None:
    response = app_client.get(endpoint, headers=headers)
    assert response.status_code == expected_status



from ._func import permission
from ._lazy import LazyPermission, lazy
from ._permissions import AllPermissions, AnyPermissions, NotPermission, Permission

__all__ = [
    "AllPermissions",
    "AnyPermissions",
    "LazyPermission",
    "NotPermission",
    "Permission",
    "lazy",
    "permission",
]



from collections.abc import Iterator
from dataclasses import dataclass

import pytest
from fastapi import Depends, FastAPI, Request, status
from fastapi.testclient import TestClient

from fastapi_has_permissions import Permission

app = FastAPI()


class HasAuthorizationHeader(Permission):
    async def check_permissions(self, request: Request) -> bool:
        return "Authorization" in request.headers


@dataclass
class HasRole(Permission):
    role: str

    async def check_permissions(self, request: Request) -> bool:
        return request.headers.get("role") == self.role


@app.get(
    "/simple-test",
    dependencies=[
        Depends(HasAuthorizationHeader()),
    ],
)
@app.get(
    "/and-test",
    dependencies=[
        Depends(HasAuthorizationHeader() & HasRole("admin")),
    ],
)
@app.get(
    "/or-test",
    dependencies=[
        Depends(HasAuthorizationHeader() | HasRole("admin")),
    ],
)
@app.get(
    "/not-test",
    dependencies=[
        Depends(~HasAuthorizationHeader()),
    ],
)
@app.get(
    "/complex-test",
    dependencies=[
        Depends((HasAuthorizationHeader() & HasRole("admin")) | ~HasRole("user")),
    ],
)
async def route() -> str:
    return "You have access to this endpoint!"


@pytest.fixture(scope="session")
def app_client() -> Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


@pytest.mark.parametrize(
    ("endpoint", "headers", "expected_status"),
    [
        pytest.param(
            "/simple-test",
            {"Authorization": "some-token"},
            status.HTTP_200_OK,
            id="simple-test-pass",
        ),
        pytest.param(
            "/simple-test",
            {},
            status.HTTP_403_FORBIDDEN,
            id="simple-test-fail-no-auth-header",
        ),
        pytest.param(
            "/and-test",
            {"Authorization": "some-token", "role": "admin"},
            status.HTTP_200_OK,
            id="and-test-pass-both-satisfied",
        ),
        pytest.param(
            "/and-test",
            {"Authorization": "some-token"},
            status.HTTP_403_FORBIDDEN,
            id="and-test-fail-missing-role",
        ),
        pytest.param(
            "/and-test",
            {"role": "admin"},
            status.HTTP_403_FORBIDDEN,
            id="and-test-fail-missing-auth",
        ),
        pytest.param(
            "/and-test",
            {},
            status.HTTP_403_FORBIDDEN,
            id="and-test-fail-neither-satisfied",
        ),
        pytest.param(
            "/and-test",
            {"Authorization": "some-token", "role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="and-test-fail-wrong-role",
        ),
        pytest.param(
            "/or-test",
            {"Authorization": "some-token", "role": "admin"},
            status.HTTP_200_OK,
            id="or-test-pass-both-satisfied",
        ),
        pytest.param(
            "/or-test",
            {"Authorization": "some-token"},
            status.HTTP_200_OK,
            id="or-test-pass-only-auth",
        ),
        pytest.param(
            "/or-test",
            {"role": "admin"},
            status.HTTP_200_OK,
            id="or-test-pass-only-role",
        ),
        pytest.param(
            "/or-test",
            {},
            status.HTTP_403_FORBIDDEN,
            id="or-test-fail-neither-satisfied",
        ),
        pytest.param(
            "/or-test",
            {"role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="or-test-fail-wrong-role-no-auth",
        ),
        pytest.param(
            "/not-test",
            {},
            status.HTTP_200_OK,
            id="not-test-pass-no-auth-header",
        ),
        pytest.param(
            "/not-test",
            {"Authorization": "some-token"},
            status.HTTP_403_FORBIDDEN,
            id="not-test-fail-auth-header-present",
        ),
        pytest.param(
            "/complex-test",
            {"Authorization": "some-token", "role": "admin"},
            status.HTTP_200_OK,
            id="complex-test-pass-auth-and-admin",
        ),
        pytest.param(
            "/complex-test",
            {},
            status.HTTP_200_OK,
            id="complex-test-pass-no-role-satisfies-not-user",
        ),
        pytest.param(
            "/complex-test",
            {"role": "guest"},
            status.HTTP_200_OK,
            id="complex-test-pass-non-user-role-satisfies-not-user",
        ),
        pytest.param(
            "/complex-test",
            {"role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="complex-test-fail-user-role-without-auth",
        ),
        pytest.param(
            "/complex-test",
            {"Authorization": "some-token", "role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="complex-test-fail-auth-with-user-role-not-admin",
        ),
    ],
)
def test_permissions(endpoint, headers, expected_status, app_client) -> None:
    response = app_client.get(endpoint, headers=headers)
    assert response.status_code == expected_status



import inspect
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from typing import Any, TypeVar, overload

from fastapi.dependencies.utils import get_typed_signature

from ._permissions import Permission
from ._results import CheckResult
from .types import AsyncFunc

TAsyncFunc = TypeVar("TAsyncFunc", bound=AsyncFunc)


@dataclass
class FuncPermission(Permission):
    func: AsyncFunc

    def __check_signature__(self) -> inspect.Signature:
        return get_typed_signature(self.func)

    async def check_permissions(self, *args: Any, **kwargs: Any) -> CheckResult:
        return await self.func(*args, **kwargs)


@overload
def permission(arg: TAsyncFunc, /) -> FuncPermission:
    pass


@overload
def permission(
    arg: None = None,
    /,
    *,
    message: str | None = None,
    status_code: int | None = None,
) -> Callable[[TAsyncFunc], FuncPermission]:
    pass


def permission(
    arg: TAsyncFunc | None = None,
    /,
    *,
    message: str | None = None,
    status_code: int | None = None,
) -> Any:
    if arg is None:
        return partial(permission, message=message, status_code=status_code)

    return FuncPermission(func=arg, message=message, status_code=status_code)


__all__ = [
    "permission",
]



from __future__ import annotations

import inspect
from collections.abc import Iterable, Sequence
from copy import copy
from dataclasses import dataclass
from typing import Annotated, Any, ClassVar, final

from fastapi import Depends
from typing_extensions import Self

from ._bases import IdentityHashMixin, SignatureOverride
from ._deps_args import (
    deps_from_func_signature,
    get_dep_arg_name,
    get_signature_with_deps,
    remap_deps_args,
    signature_with_params,
)
from ._errors import HTTPExcRaiser
from ._resolvers import PermissionResolver, Resolvable, ResolvedPermission
from ._results import CheckResult, Failed, Skipped, is_skipped
from .types import AsyncFunc, Dep


async def _default_check_permissions(_: Any) -> CheckResult:
    return True


@dataclass
class Permission(HTTPExcRaiser, Resolvable, SignatureOverride, IdentityHashMixin):
    check_permissions: ClassVar[AsyncFunc[CheckResult]] = _default_check_permissions

    def __deps__(self) -> Iterable[Dep]:
        for name in deps_from_func_signature(self.__init__):
            yield getattr(self, name)

    def __get_signature__(self) -> inspect.Signature:
        return signature_with_params([self.__to_sign_param__()])

    def __check_signature__(self) -> inspect.Signature:
        return get_signature_with_deps(self.check_permissions, [*self.__deps__()])

    def __to_sign_param__(self, idx: int = 0, /) -> inspect.Parameter:
        return inspect.Parameter(
            name=get_dep_arg_name(idx),
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=Annotated[
                PermissionResolver,
                self.__resolver_to_depends__(self.__to_resolver__()),
            ],
        )

    def __to_resolver__(self) -> PermissionResolver:
        return PermissionResolver(permission=self)

    def __resolver_to_depends__(self, resolver: PermissionResolver) -> Any:
        return Depends(resolver)

    @remap_deps_args
    async def __call__(self, permission: ResolvedPermission) -> None:
        message: str | None = None
        result: bool

        match await permission.check_permissions():
            case Failed(reason=reason):
                message = reason
                result = False
            case False:
                message = self.get_exc_message()
                result = False
            case _:
                result = True

        if not result:
            self.raise_http_exception(message)

    def __and__(self, other: Permission) -> Permission:
        if not isinstance(other, Permission):
            return NotImplemented

        return AllPermissions([self, other])

    def __or__(self, other: Permission) -> Permission:
        if not isinstance(other, Permission):
            return NotImplemented

        return AnyPermissions([self, other])

    def __invert__(self) -> Permission:
        return NotPermission(self)


@dataclass
class _AllAnyPermissions(Permission):
    permissions: Sequence[Permission]

    def __check_signature__(self) -> inspect.Signature:
        return signature_with_params([permission.__to_sign_param__(i) for i, permission in enumerate(self.permissions)])

    def _merge_permissions(self, other: Permission, cls: type[Self]) -> Self:
        new_self = copy(self)

        if isinstance(other, cls):
            new_self.permissions = [*self.permissions, *other.permissions]
        else:
            new_self.permissions = [*self.permissions, other]

        return new_self


@dataclass
class PermissionWrapper(Permission):
    permission: Permission

    def __check_signature__(self) -> inspect.Signature:
        return signature_with_params([self.permission.__to_sign_param__()])

    def __deps__(self) -> Iterable[Dep]:
        return []

    @remap_deps_args
    async def check_permissions(self, permission: ResolvedPermission) -> CheckResult:
        return await permission.check_permissions()


@final
class AnyPermissions(_AllAnyPermissions):
    default_exc_message: ClassVar[str] = "None of the permissions were satisfied"

    def __or__(self, other: Permission) -> Permission:
        if not isinstance(other, Permission):
            return NotImplemented

        return self._merge_permissions(other, AnyPermissions)

    async def check_permissions(self, *permissions: ResolvedPermission) -> CheckResult:
        only_skips = False

        for permission in permissions:
            result = await permission.check_permissions()

            if is_skipped(result):
                only_skips = True
                continue

            if result:
                return True

        if only_skips:
            return Skipped()

        return False


@final
@dataclass
class AllPermissions(_AllAnyPermissions):
    default_exc_message: ClassVar[str] = "Not all permissions were satisfied"

    def __and__(self, other: Permission) -> Permission:
        if not isinstance(other, Permission):
            return NotImplemented

        return self._merge_permissions(other, AllPermissions)

    async def check_permissions(self, *permissions: ResolvedPermission) -> CheckResult:
        only_skips = False

        for permission in permissions:
            result = await permission.check_permissions()

            if is_skipped(result):
                only_skips = True
                continue

            if not result:
                return False

        if only_skips:
            return Skipped()

        return True


@final
@dataclass
class NotPermission(PermissionWrapper):
    default_exc_message: ClassVar[str] = "The permission was satisfied, but it should not have been"

    async def check_permissions(self, permission: ResolvedPermission) -> CheckResult:
        result = await permission.check_permissions()

        if is_skipped(result):
            return result

        return not result

    def __invert__(self) -> Permission:
        return self.permission


__all__ = [
    "AllPermissions",
    "AnyPermissions",
    "NotPermission",
    "Permission",
    "PermissionWrapper",
]



from collections.abc import Iterator
from typing import Annotated

import pytest
from fastapi import Depends, FastAPI, Header, status
from fastapi.testclient import TestClient
from starlette.requests import Request

from fastapi_has_permissions import Permission
from fastapi_has_permissions._func import FuncPermission, permission
from fastapi_has_permissions._results import CheckResult

app = FastAPI()


async def has_authorization_header(request: Request) -> CheckResult:
    return "authorization" in request.headers


async def has_role(role: Annotated[str, Header()]) -> CheckResult:
    return role == "admin"


has_authorization = permission(has_authorization_header)
has_admin_role = permission(has_role)


@app.get(
    "/simple-test",
    dependencies=[Depends(has_authorization)],
)
@app.get(
    "/and-test",
    dependencies=[Depends(has_authorization & has_admin_role)],
)
@app.get(
    "/or-test",
    dependencies=[Depends(has_authorization | has_admin_role)],
)
@app.get(
    "/not-test",
    dependencies=[Depends(~has_authorization)],
)
@app.get(
    "/complex-test",
    dependencies=[Depends((has_authorization & has_admin_role) | ~has_authorization)],
)
async def route():
    return "OK"


@pytest.fixture(scope="session")
def app_client() -> Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


@pytest.mark.parametrize(
    ("endpoint", "headers", "expected_status"),
    [
        pytest.param(
            "/simple-test",
            {"authorization": "Bearer token"},
            status.HTTP_200_OK,
            id="simple-pass",
        ),
        pytest.param(
            "/simple-test",
            {},
            status.HTTP_403_FORBIDDEN,
            id="simple-fail-missing-auth",
        ),
        pytest.param(
            "/and-test",
            {"authorization": "Bearer token", "role": "admin"},
            status.HTTP_200_OK,
            id="and-test-pass",
        ),
        pytest.param(
            "/and-test",
            {"authorization": "Bearer token", "role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="and-test-fail-wrong-role",
        ),
        pytest.param(
            "/and-test",
            {"role": "admin"},
            status.HTTP_403_FORBIDDEN,
            id="and-test-fail-missing-auth",
        ),
        pytest.param(
            "/and-test",
            {},
            status.HTTP_422_UNPROCESSABLE_ENTITY,
            id="and-test-fail-missing-both",
        ),
        pytest.param(
            "/or-test",
            {"authorization": "Bearer token"},
            status.HTTP_422_UNPROCESSABLE_ENTITY,
            id="or-test-fail-missing-role-header",
        ),
        pytest.param(
            "/or-test",
            {"role": "admin"},
            status.HTTP_200_OK,
            id="or-test-pass-role-only",
        ),
        pytest.param(
            "/or-test",
            {"authorization": "Bearer token", "role": "admin"},
            status.HTTP_200_OK,
            id="or-test-pass-both",
        ),
        pytest.param(
            "/or-test",
            {"role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="or-test-fail-wrong-role-no-auth",
        ),
        pytest.param(
            "/not-test",
            {},
            status.HTTP_200_OK,
            id="not-test-pass-no-auth",
        ),
        pytest.param(
            "/not-test",
            {"authorization": "Bearer token"},
            status.HTTP_403_FORBIDDEN,
            id="not-test-fail-has-auth",
        ),
        pytest.param(
            "/complex-test",
            {"authorization": "Bearer token", "role": "admin"},
            status.HTTP_200_OK,
            id="complex-pass-auth-and-admin",
        ),
        pytest.param(
            "/complex-test",
            {},
            status.HTTP_422_UNPROCESSABLE_ENTITY,
            id="complex-fail-missing-role-header",
        ),
        pytest.param(
            "/complex-test",
            {"authorization": "Bearer token", "role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="complex-fail-auth-wrong-role",
        ),
    ],
)
def test_permissions(endpoint, headers, expected_status, app_client) -> None:
    response = app_client.get(endpoint, headers=headers)
    assert response.status_code == expected_status


def test_permission_is_permission_subclass() -> None:
    assert isinstance(has_authorization, Permission)
    assert isinstance(has_admin_role, Permission)


def test_func_permission_stores_func() -> None:
    assert isinstance(has_authorization, FuncPermission)
    assert has_authorization.func is has_authorization_header

    assert isinstance(has_admin_role, FuncPermission)
    assert has_admin_role.func is has_role



from collections.abc import Iterator

import pytest
from fastapi import Depends, FastAPI, Request, status
from fastapi.testclient import TestClient

from fastapi_has_permissions import Permission
from fastapi_has_permissions._results import SkipPermissionCheck

app = FastAPI()


class AlwaysSkip(Permission):
    async def check_permissions(self) -> bool:
        raise SkipPermissionCheck(reason="always skip")


class HasAuthorizationHeader(Permission):
    async def check_permissions(self, request: Request) -> bool:
        return "Authorization" in request.headers


class SkipIfNoToken(Permission):
    """Skips if no Authorization header, otherwise checks if token is valid."""

    async def check_permissions(self, request: Request) -> bool:
        if "Authorization" not in request.headers:
            raise SkipPermissionCheck(reason="no token provided")

        return request.headers["Authorization"] == "valid-token"


class HasRole(Permission):
    async def check_permissions(self, request: Request) -> bool:
        return request.headers.get("role") == "admin"


@app.get(
    "/simple-skip",
    dependencies=[
        Depends(AlwaysSkip()),
    ],
)
@app.get(
    "/skip-if-no-token",
    dependencies=[
        Depends(SkipIfNoToken()),
    ],
)
@app.get(
    "/and-skip-and-pass",
    dependencies=[
        Depends(AlwaysSkip() & HasAuthorizationHeader()),
    ],
)
@app.get(
    "/and-skip-and-fail",
    dependencies=[
        Depends(AlwaysSkip() & HasRole()),
    ],
)
@app.get(
    "/and-all-skip",
    dependencies=[
        Depends(AlwaysSkip() & AlwaysSkip()),
    ],
)
@app.get(
    "/or-skip-and-pass",
    dependencies=[
        Depends(AlwaysSkip() | HasAuthorizationHeader()),
    ],
)
@app.get(
    "/or-skip-and-fail",
    dependencies=[
        Depends(AlwaysSkip() | HasRole()),
    ],
)
@app.get(
    "/or-all-skip",
    dependencies=[
        Depends(AlwaysSkip() | AlwaysSkip()),
    ],
)
@app.get(
    "/not-skip",
    dependencies=[
        Depends(~AlwaysSkip()),
    ],
)
@app.get(
    "/complex-skip",
    dependencies=[
        Depends((SkipIfNoToken() & HasRole()) | HasAuthorizationHeader()),
    ],
)
async def route() -> str:
    return "You have access to this endpoint!"


@pytest.fixture(scope="session")
def app_client() -> Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


@pytest.mark.parametrize(
    ("endpoint", "headers", "expected_status"),
    [
        pytest.param(
            "/simple-skip",
            {},
            status.HTTP_200_OK,
            id="simple-skip-always-skipped",
        ),
        pytest.param(
            "/skip-if-no-token",
            {},
            status.HTTP_200_OK,
            id="skip-if-no-token-missing-header-skips",
        ),
        pytest.param(
            "/skip-if-no-token",
            {"Authorization": "valid-token"},
            status.HTTP_200_OK,
            id="skip-if-no-token-valid-token-passes",
        ),
        pytest.param(
            "/skip-if-no-token",
            {"Authorization": "invalid-token"},
            status.HTTP_403_FORBIDDEN,
            id="skip-if-no-token-invalid-token-fails",
        ),
        pytest.param(
            "/and-skip-and-pass",
            {"Authorization": "some-token"},
            status.HTTP_200_OK,
            id="and-skip-and-pass-auth-present",
        ),
        pytest.param(
            "/and-skip-and-fail",
            {},
            status.HTTP_403_FORBIDDEN,
            id="and-skip-and-fail-no-role",
        ),
        pytest.param(
            "/and-skip-and-fail",
            {"role": "admin"},
            status.HTTP_200_OK,
            id="and-skip-and-fail-has-admin-role",
        ),
        pytest.param(
            "/and-skip-and-fail",
            {"role": "user"},
            status.HTTP_403_FORBIDDEN,
            id="and-skip-and-fail-wrong-role",
        ),
        pytest.param(
            "/and-all-skip",
            {},
            status.HTTP_200_OK,
            id="and-all-skip-both-skipped",
        ),
        pytest.param(
            "/or-skip-and-pass",
            {"Authorization": "some-token"},
            status.HTTP_200_OK,
            id="or-skip-and-pass-auth-present",
        ),
        pytest.param(
            "/or-skip-and-fail",
            {},
            status.HTTP_200_OK,
            id="or-skip-and-fail-no-role-skipped",
        ),
        pytest.param(
            "/or-skip-and-fail",
            {"role": "user"},
            status.HTTP_200_OK,
            id="or-skip-and-fail-wrong-role-skipped",
        ),
        pytest.param(
            "/or-all-skip",
            {},
            status.HTTP_200_OK,
            id="or-all-skip-both-skipped",
        ),
        pytest.param(
            "/not-skip",
            {},
            status.HTTP_200_OK,
            id="not-skip-passthrough",
        ),
        pytest.param(
            "/complex-skip",
            {},
            status.HTTP_403_FORBIDDEN,
            id="complex-skip-no-headers",
        ),
        pytest.param(
            "/complex-skip",
            {"Authorization": "valid-token"},
            status.HTTP_200_OK,
            id="complex-skip-valid-token-no-role",
        ),
        pytest.param(
            "/complex-skip",
            {"Authorization": "valid-token", "role": "admin"},
            status.HTTP_200_OK,
            id="complex-skip-valid-token-and-admin-role",
        ),
        pytest.param(
            "/complex-skip",
            {"Authorization": "invalid-token", "role": "admin"},
            status.HTTP_200_OK,
            id="complex-skip-invalid-token-but-has-auth-header",
        ),
        pytest.param(
            "/complex-skip",
            {"role": "admin"},
            status.HTTP_200_OK,
            id="complex-skip-admin-role-no-auth-skips-token-check",
        ),
    ],
)
def test_permissions(endpoint, headers, expected_status, app_client) -> None:
    response = app_client.get(endpoint, headers=headers)
    assert response.status_code == expected_status



from __future__ import annotations

import inspect
from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias

from ._bases import IdentityHashMixin, SignatureOverride
from ._deps_args import remap_deps_args
from ._results import CheckResult, Skipped, call_permissions_check
from .types import Args, Kwargs

if TYPE_CHECKING:
    from ._permissions import Permission


@dataclass
class ResolvedPermission(IdentityHashMixin):
    permission: Permission
    args: Args
    kwargs: Kwargs

    async def check_permissions(self) -> CheckResult:
        return await call_permissions_check(self.permission, *self.args, **self.kwargs)


@dataclass
class ResolvedToSkipped(IdentityHashMixin):
    reason: str | None = None

    async def check_permissions(self) -> Skipped:
        return Skipped(reason=self.reason)


ResolvedResult: TypeAlias = ResolvedPermission | ResolvedToSkipped


@dataclass
class PermissionResolver(IdentityHashMixin, SignatureOverride):
    permission: Permission

    def __get_signature__(self) -> inspect.Signature:
        return self.permission.__check_signature__()

    @remap_deps_args
    async def __call__(self, *args: Any, **kwargs: Any) -> ResolvedResult:
        return ResolvedPermission(self.permission, args, kwargs)


class Resolvable(Protocol):
    @abstractmethod
    def __to_resolver__(self) -> PermissionResolver:
        pass


__all__ = [
    "PermissionResolver",
    "Resolvable",
    "ResolvedPermission",
    "ResolvedResult",
    "ResolvedToSkipped",
]



from collections.abc import Callable, Coroutine, Sequence
from typing import Any, Literal, TypeAlias

from typing_extensions import TypeVar

_TAny = TypeVar("_TAny", default=Any)

Args: TypeAlias = tuple[_TAny, ...]
Kwargs: TypeAlias = dict[str, _TAny]

# TODO: Add typing for dependency input param
Dep: TypeAlias = Any
Deps: TypeAlias = Sequence[Any]

DepScope: TypeAlias = Literal["function", "request"]

Func: TypeAlias = Callable[..., _TAny]
AsyncFunc: TypeAlias = Callable[..., Coroutine[Any, Any, _TAny]]

Exceptions: TypeAlias = tuple[type[BaseException], ...]

__all__ = [
    "Args",
    "AsyncFunc",
    "Dep",
    "DepScope",
    "Deps",
    "Exceptions",
    "Func",
    "Kwargs",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar

from typing_extensions import TypeIs

from .types import AsyncFunc

if TYPE_CHECKING:
    from ._permissions import Permission


class SkipPermissionCheck(Exception):  # noqa: N818
    def __init__(self, reason: str | None = None) -> None:
        super().__init__(reason)
        self.reason = reason


class PermissionCheckFailed(Exception):  # noqa: N818
    def __init__(self, reason: str | None = None) -> None:
        super().__init__(reason)
        self.reason = reason


@dataclass
class Skipped:
    reason: str | None = None


@dataclass
class Failed:
    reason: str | None = None

    def __bool__(self) -> bool:
        return False


CheckResult: TypeAlias = bool | Skipped | Failed


def is_skipped(result: CheckResult) -> TypeIs[Skipped]:
    return isinstance(result, Skipped)


def is_failed(result: CheckResult) -> TypeIs[Failed]:
    return isinstance(result, Failed)


TAsyncFunc = TypeVar("TAsyncFunc", bound=AsyncFunc)


async def call_permissions_check(
    permission: Permission,
    /,
    *args: Any,
    **kwargs: Any,
) -> CheckResult:
    try:
        result = await permission.check_permissions(*args, **kwargs)
    except PermissionCheckFailed as exc:
        return Failed(reason=exc.reason or permission.get_exc_message())
    except SkipPermissionCheck as exc:
        return Skipped(reason=exc.reason)

    match result:
        case False:
            return Failed(reason=permission.get_exc_message())
        case _:
            return result


__all__ = [
    "CheckResult",
    "Failed",
    "SkipPermissionCheck",
    "Skipped",
    "call_permissions_check",
    "is_failed",
    "is_skipped",
]



import inspect
from collections.abc import Iterable
from functools import wraps
from itertools import count, takewhile
from typing import Annotated, Any, TypeVar, cast

from fastapi.dependencies.utils import get_typed_signature
from fastapi.params import Depends

from .types import AsyncFunc, Dep, Deps, Func


def get_dep_arg_name(index: int = 0, /) -> str:
    return f"__positional_arg_{index}__"


def _get_dep_annotation(dep: Any, /) -> Any:
    if isinstance(dep, Depends):
        return Annotated[Any, dep]

    return dep


def get_signature_with_deps(func: Func, deps: Deps) -> inspect.Signature:
    sign = get_typed_signature(func)

    # skip the self + first N parameters that correspond to parametrized dependencies
    other_deps = [*sign.parameters.values()][len(deps) :]

    return sign.replace(
        parameters=[
            *[
                inspect.Parameter(
                    name=get_dep_arg_name(i),
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=_get_dep_annotation(dep),
                )
                for i, dep in enumerate(deps)
            ],
            *other_deps,
        ],
    )


def deps_from_func_signature(func: Func) -> Iterable[str]:
    sign = get_typed_signature(func)

    for param in sign.parameters.values():
        if param.annotation is Dep:
            yield param.name


def signature_with_params(params: Iterable[inspect.Parameter]) -> inspect.Signature:
    return inspect.Signature(parameters=[*params])


TAsyncFunc = TypeVar("TAsyncFunc", bound=AsyncFunc)


def remap_deps_args(func: TAsyncFunc) -> TAsyncFunc:
    @wraps(func)
    async def wrapper(self: Any, **kwargs: Any) -> Any:
        possible_args = (get_dep_arg_name(i) for i in count())
        args = [kwargs.pop(arg) for arg in takewhile(lambda key: key in kwargs, possible_args)]
        return await func(self, *args, **kwargs)

    return cast("TAsyncFunc", wrapper)


__all__ = [
    "deps_from_func_signature",
    "get_dep_arg_name",
    "get_signature_with_deps",
    "remap_deps_args",
    "signature_with_params",
]



from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

from fastapi import Security
from fastapi.security import SecurityScopes

from ._permissions import Permission
from ._resolvers import PermissionResolver
from .types import Dep


@dataclass
class IsAuthenticated(Permission):
    authentication_dep: Dep

    async def check_permissions(self, is_authenticated: bool, /) -> bool:  # noqa: FBT001
        return is_authenticated


@dataclass
class HasScope(Permission):
    scope_dep: Dep
    scopes: Iterable[str]

    def __resolver_to_depends__(self, resolver: PermissionResolver) -> Any:
        return Security(resolver, scopes=[*self.scopes])

    async def check_permissions(self, current_scopes: Iterable[str], /, security_scopes: SecurityScopes) -> bool:
        return all(required_scope in current_scopes for required_scope in security_scopes.scopes)


@dataclass
class HasRole(Permission):
    role_dep: Dep
    roles: Iterable[str]

    async def check_permissions(self, current_role: str, /) -> bool:
        return current_role in self.roles


__all__ = [
    "HasRole",
    "HasScope",
    "IsAuthenticated",
]



import inspect
from contextlib import suppress
from dataclasses import dataclass, field
from functools import cached_property, lru_cache, partial
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias, TypeVar, cast, overload

from fastapi import BackgroundTasks, Request, Response, WebSocket, routing
from fastapi.dependencies import utils
from fastapi.dependencies.utils import get_typed_signature
from fastapi.exceptions import RequestValidationError

from ._permissions import Permission, PermissionWrapper
from ._resolvers import PermissionResolver, ResolvedResult, ResolvedToSkipped
from .types import Exceptions

_DEPENDENCY_CACHE_KEY = "__fastapi_has_permissions_dependency_cache__"

if not TYPE_CHECKING:
    _base_solve_dependencies = utils.solve_dependencies

    async def _solve_dependencies(
        *,
        request: Request | WebSocket,
        **kwargs: Any,
    ) -> utils.SolvedDependency:
        solved = await _base_solve_dependencies(request=request, **kwargs)

        # cache the solved dependencies for the current request to avoid redundant computations
        request.scope[_DEPENDENCY_CACHE_KEY] = solved.dependency_cache

        return solved

    utils.solve_dependencies = _solve_dependencies
    routing.solve_dependencies = _solve_dependencies

_AnyRoute: TypeAlias = routing.APIRoute | routing.APIWebSocketRoute


@lru_cache(maxsize=1024)
def _resolver_to_dependant(path: str, resolver: PermissionResolver) -> utils.Dependant:
    return utils.get_dependant(
        path=path,
        call=resolver,
    )


async def _get_request_body(request: Request) -> Any:
    # TODO: add proper handling of json/body
    with suppress(Exception):
        return await request.json()

    with suppress(Exception):
        return await request.body()

    return None


async def _solve_dependencies_for_dependant(
    request: Request,
    response: Response,
    dependant: utils.Dependant,
) -> utils.SolvedDependency:
    route: _AnyRoute = request.scope["route"]

    dependency_cache = request.scope.get(_DEPENDENCY_CACHE_KEY, {})
    body = await _get_request_body(request)

    return await utils.solve_dependencies(
        request=request,
        dependant=dependant,
        body=body,
        dependency_cache=dependency_cache,
        background_tasks=cast("BackgroundTasks", response.background),
        response=response,
        dependency_overrides_provider=route.dependency_overrides_provider,
        # this param is no longer actually used, can ignore it
        async_exit_stack=None,  # type: ignore[arg-type]
        embed_body_fields=route._embed_body_fields,  # noqa: SLF001
    )


@dataclass
class LazyPermissionResolver(PermissionResolver):
    skip_on_exc: Exceptions = field(default=(), kw_only=True)

    def __get_signature__(self) -> inspect.Signature:
        return get_typed_signature(self.__call__)

    @cached_property
    def _eager_resolver(self) -> PermissionResolver:
        return PermissionResolver(self.permission)

    async def __call__(self, request: Request, response: Response) -> ResolvedResult:
        route: _AnyRoute = request.scope["route"]
        dependant = _resolver_to_dependant(route.path, self._eager_resolver)

        try:
            solved = await _solve_dependencies_for_dependant(request, response, dependant)

            if solved.errors:
                raise RequestValidationError(solved.errors)
        except self.skip_on_exc:
            return ResolvedToSkipped()

        return await self._eager_resolver(**solved.values)


@dataclass
class _HasSkipOnExc:
    skip_on_exc: Exceptions = field(default=(), kw_only=True)


@dataclass
class LazyPermission(_HasSkipOnExc, Permission):
    def __to_resolver__(self) -> LazyPermissionResolver:
        return LazyPermissionResolver(self, skip_on_exc=self.skip_on_exc)


@dataclass
class LazyPermissionWrapper(_HasSkipOnExc, PermissionWrapper):
    def __to_resolver__(self) -> LazyPermissionResolver:
        return LazyPermissionResolver(self.permission, skip_on_exc=self.skip_on_exc)


TCls = TypeVar("TCls", bound=type[Permission])
TPermission = TypeVar("TPermission", bound=Permission)


class _LazyDecorator(Protocol):
    @overload
    def __call__(self, arg: TCls, /) -> TCls:
        pass

    @overload
    def __call__(self, arg: TPermission, /) -> LazyPermissionWrapper:
        pass


@overload
def lazy(
    arg: None = None,
    /,
    *,
    skip_on_exc: Exceptions | None = None,
) -> _LazyDecorator:
    pass


@overload
def lazy(cls: TCls, /) -> TCls:
    pass


@overload
def lazy(
    permission: TPermission,
    /,
    *,
    skip_on_exc: Exceptions | None = None,
) -> LazyPermissionWrapper:
    pass


def lazy(
    arg: type[Permission] | Permission | None = None,
    /,
    *,
    skip_on_exc: Exceptions | None = None,
) -> type[Permission] | LazyPermissionWrapper | _LazyDecorator:
    if arg is None:
        return cast(
            "_LazyDecorator",
            partial(lazy, skip_on_exc=skip_on_exc),
        )

    if isinstance(arg, type):
        if issubclass(arg, LazyPermission):
            raise TypeError("Cannot apply @lazy to a subclass of LazyPermission")

        ns = {}
        if skip_on_exc is not None:
            ns["skip_on_exc"] = skip_on_exc

        return type(f"Lazy{arg.__name__}", (LazyPermission, arg), ns)

    return LazyPermissionWrapper(
        permission=arg,
        skip_on_exc=skip_on_exc or (),
    )


__all__ = [
    "LazyPermission",
    "LazyPermissionResolver",
    "lazy",
]
